/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef BPRED_H
#define BPRED_H

#include <stdio.h>
#include "ss.h"

enum bpred_class {
  BPred2bitBTB,			/* 2-bit saturating counter predictor */
  BPredTaken,			/* static predict taken */
  BPredNotTaken,		/* static predict not taken */
};

struct bpred_btb_ent {
  SS_ADDR_TYPE addr;		/* address of branch being tracked */
  unsigned char state;		/* 0, 1 -> not taken /// 2, 3, taken */
  SS_ADDR_TYPE target;		/* last destination when taken */
};

struct bpred {
  enum bpred_class class;	/* type of predictor */
  SS_COUNTER_TYPE hits;		/* num correct predictions */
  SS_COUNTER_TYPE misses;	/* num incorrect predictions */
  union {
    struct {
      unsigned long size;	/* number of entries in direct-mapped BTB */
      struct bpred_btb_ent *sets;/* BTB set table */
    } btb;
  } data;
};

/* create a branch predictor */
struct bpred *bpred_create(enum bpred_class class, unsigned long size);

/* print branch predictor configuration */
void bpred_config(struct bpred *pred, FILE *stream);

/* print predictor stats */
void bpred_stats(struct bpred *pred, FILE *stream);

SS_ADDR_TYPE bpred_lookup(struct bpred *pred,
			  SS_ADDR_TYPE baddr, SS_ADDR_TYPE btarget);

/* update the branch predictor, only useful for stateful predictors;
   updates entry for BADDR, branch jumped to BTARGET and was taken
   if TAKEN is non-zero, predictor statistics are updated with result
   of prediction, indicated by CORRECT */
void bpred_update(struct bpred *pred, SS_ADDR_TYPE baddr,
		  SS_ADDR_TYPE btarget, int taken, int correct);
void bpred_dump(struct bpred *pred, FILE *stream);

#endif /* BPRED_H */
